/**************************************************************************
 *  Copyright (c) 2002  Hartmut Ring, Universitaet Siegen                 *
 *  http://www.math.uni-siegen.de/ring                                    *
 *                                                                        *
 *  This file is part of the pariGUIde package.                           *
 *                                                                        *
 *  pariGUIde is free software; you can redistribute it and/or modify it  *
 *  under the terms of the GNU General Public License (GPL)               *
 *  as published by the Free Software Foundation:                         *
 *  http://www.opensource.org/licenses/gpl-license.html                   *
 *************************************************************************/
// editwindow.cpp

#include "editwindow.h"
#include "gpcontroller.h"
#include "pariGUIwindow.h"
#include "matrixdlg.h"
#include "sumdlg.h"
#include "proddlg.h"
#include "tabledlg.h"
#include "plotdlg.h"
#include "intdlg.h"

#include <qapplication.h>
#include <qstatusbar.h>

#include <qcheckbox.h>
#include <qfontdialog.h>
#include <qfiledialog.h>
#include <qinputdialog.h>
#include <qtoolbutton.h>
#include <qmessagebox.h>
#include <qtextcodec.h>
#include <qcombobox.h>

#include <qstyle.h>

#define INPUT_ECHO_COLOR    "\"red\""
#define ERROR_COLOR         "\"#C000C0\""
#define MATRIX_COLOR        "\"#E8FFEC\""
#define TABLE_HEADING_COLOR "\"#C8CCFF\""
#define TABLE_FIELD_COLOR   "\"#E8ECFF\""

#ifdef _WIN32
#define PARI_PROMPT "~" // that must defined in pari.cfg (only WIN)!
#endif

#ifdef _LINUX
#define PARI_PROMPT "?"
#endif

EditWindow::EditWindow (QWidget* parent, GpController* pariGP)
: QSplitter (parent),
  _pari (pariGP),
  _firstInput (true),
  _realPrecision (28),
  _font ("Courier", 14)
{
  setOrientation (QSplitter::Vertical);

  _tabView = new MyQTabWidget (this);
  _tabView->setGeometry (0, 0, 780, 480);
  //_tabView->setMargin (0);

  _recordWnd = new QTextView (_tabView);
  _recordWnd->setTextFormat (RichText);

  _recordWnd->setText (tr(
    "<h1><font size=\"+8\"><font color=\"#808080\">pari</font>GUI</font>de</h1>"
    "<h2>version " PARI_GUI_VERSION "<br>"
    "Autor and Copyright &copy; 2002 &nbsp;Hartmut Ring</h2>"
    "<p>Linux Port: Henning Westerholt</p>"
    "<p>Preparation of the Pari-GP Documentation: Nils-Peter Skoruppa</p>"

    //"<table bgcolor=\"white\" cellspacing=\"0\" cellpadding=\"4\" border=\"1\"><tr><td>"
    "<p>This program use <b>GP/PARI CALCULATOR</b> as server.<br>"
    "Copyright &copy; C. Batut, K. Belabas, D. Bernardi, H. Cohen and M. Olivier</p>"
    //"</td></tr></table>"

    "<p><font color=\"#008040\">pariGUIde is free software; you can redistribute it and/or modify it "
    "under the terms of the GNU General Public License (GPL) as published by the Free Software Foundation:"
    "http://www.opensource.org/licenses/gpl-license.html.<br></font></p>"
    "<br>"



    //"<table bgcolor=" MATRIX_COLOR " cellspacing=\"0\" cellpadding=\"10\" border=\"0\"><tr><td>"
    "<font size=\"+1\"><b>Please type your input into the window below.<br>"
    "You can use line breaks (return).<br>"
    "Press Ctrl+Return to send.</b></font>"
    //"</td></tr></table>"
  ));

  _tutorialWnd  = new QTextBrowser (_tabView);
  _referenceWnd = new QTextBrowser (_tabView);

  _plotWnd = new PlotView (_tabView);

  QFont f ("Times", 14);
  _tutorialWnd->setFont (f);
  _referenceWnd->setFont (f);
  //_referenceWnd->mimeSourceFactory()->setExtensionType("html", "text/utf8");

#ifdef _WIN32
  QString path = _pari->programPath();
#endif

#ifdef _LINUX
  QString path = "/usr/share/pariguide/";
#endif

  QString locale = QTextCodec::locale();  //try to find the right language, currently only 'de' and 'en'
  if (locale.contains("de", TRUE) > 0)
    locale = "de";
  else locale = "en";

#ifdef _LINUX

  _tutorialWnd->mimeSourceFactory()->setFilePath  (path + "reference" );


  _referenceWnd->mimeSourceFactory()->setFilePath (path + "reference");
#endif

#ifdef _WIN32
  _tutorialWnd->mimeSourceFactory()->setFilePath  (path + "doc" );
  _referenceWnd->mimeSourceFactory()->setFilePath (path + "doc");
#endif

  _tutorialWnd->setSource ("tutor." +  locale + ".html");
  _referenceWnd->setSource ("index.html");

  _tabView->addTab (_recordWnd, tr("Protocol"));
  _tabView->addTab (_plotWnd, tr("Plot"));
  _tabView->addTab (_tutorialWnd, tr("Tutorial"));
  _tabView->addTab (_referenceWnd, tr( "Reference"));

  //_tabView->setTabPosition (QTabWidget::Bottom);
  //_tabView->setMargin(0);

  _inputWnd = new QMultiLineEdit (this);
  _inputWnd->setGeometry (0, 0, 780, 100);

  _font.setBold (true);
  _inputWnd->setFont (_font);
  _recordWnd->setFont (_font);

#if (!defined KEIN_PATCH_WEGEN_QT_FEHLER) //no patch, because of an Qt-error
  _timer = new QTimer (this);
  connect (_timer, SIGNAL(timeout()), this, SLOT(scrollDown()));
#endif

  connect (_inputWnd, SIGNAL(returnPressed()), this, SLOT(newLine()));
  _inputWnd->setFocus();
}

void EditWindow::formatMatrix (QString& s, bool table/*=false*/) {
  // without checking!
  // in the raw-mode you have three options of matrix-output (examples):
  // (a) [1,2;3,4]
  // (b) Mat([1,2,3])   (distinction to a vector)
  // (c) Mat(1)

  int depth = 0;
  for (int i=0; i<(int)s.length(); i++) {
    QChar c = s[i];
    if (depth == 0 && c == 'M') {
      if (s.mid(i,4) == "Mat(") {

        if (s[i+4] == '[') { // case (b): Mat([1,2,3]) ---> [1,2,3]
          s.remove(i,4);
          int p;
          if ((p = s.findRev(')')) >= 0)
            s.remove(p,1);
        }
        else // case (c):  Mat(1) ---> (1)
          s.remove(i,3);
        c = s[i];
      }
    }
    if (c == '[' || c == '(') depth++;

    if (depth == 1) {
      if (c == '[' || c == '(') { // begin of the table
        s.remove (i,1);
        s.insert (i, QString("<table bgcolor=")
                     + (table ? TABLE_FIELD_COLOR : MATRIX_COLOR)
                     + " cellspacing=\"0\" border=\"1\">"
                       "<tr><td align=\"right\">");
        i = s.find ('>', i);
      }
      else if (c == ']' || c == ')') { // end of the table
        s.remove (i,1);
        s.insert (i, "</td></tr></table>");
        break;
      }
      else if (c == ',') { // new element in the row
        s.remove (i,1);
        s.insert (i, "</td><td align=\"right\">");
      }
      else if (c == ';') { // new row
        s.remove (i,1);
        s.insert (i, "</td></tr><tr><td align=\"right\">");
      }
    }

    if (c == ']' || c == ')') depth--;
  }
  if (table)
    s.replace (QRegExp(">\\s*<b>"), " bgcolor=" TABLE_HEADING_COLOR "><b>");
}

void EditWindow::newLine() {
   int line, col;
  _inputWnd->getCursorPosition (&line, &col);
  if (line > 0) {
    QString s = _inputWnd->textLine(line-1);
    int i = 0;
    while (i < (int)s.length() && s[i] == ' ') {
      i++;
      _inputWnd->insert(" ");
    }
  }
}

void EditWindow::appendRecord (const QString& output) {
  _record += output;
  _tabView->showPage(_recordWnd);
  _recordWnd->setFocus();
  _recordWnd->setText (_record);
  _inputWnd->setFocus();
  scrollToEndOfRecord();
}

void EditWindow::scrollToEndOfRecord() {
#if (!defined KEIN_PATCH_WEGEN_QT_FEHLER) //no patch, because of an Qt error
  _timer->start (50, TRUE);
#else
  _recordWnd->setContentsPos (0, 0x10000);
  // If tables (matrices) appears in the window, Qt scrolls the windows
  // after end of this function back to the first matrix!!!????!!!!
#endif
}

void EditWindow::dispatch (QString pariInput, QString& output, QString& error) { //generic IO methode
  // format to single-row for PARI
  pariInput.replace (QRegExp("\\\\\\\\[^\\n]*"),"");// remove the newline-comments
  pariInput.replace (QRegExp("\\n"), "");           // connect the lines
  pariInput.replace (QRegExp("[{}]"), "");          // remove the brackets

// doesn't work yet
//  pariInput.replace (QRegExp("\\q"), "");           // prevent killing of PARI

  ((QMainWindow*)parentWidget())->statusBar()->message(tr("calculating.."
                       "to cancel press [Alt]+[Ctrl]+C (not implemented yet)"));

  pariInput += "\n1/0\n";   // only one newline at the end, then provoke a failure
  _pari->writeStdIn (pariInput);

  while ((! _pari->stdErrAvailable()) && (! GpController::isBreakHotkeyPressed())) { //doesn't work under Linux, stdErrAvailable blocks
    if (! _pari->stdErrAvailable())
      if (GpController::isBreakHotkeyPressed()) {
        if (QMessageBox::information (this, "pariGUIde", tr("stop computation?"), QMessageBox::Yes, QMessageBox::No)
            == QMessageBox::Yes)
          _pari->sendCtrlC();
      }
      if (_pari->stdOutAvailable())
        output += _pari->readStdOut();
  }

  while (_pari->stdErrAvailable())
    error += _pari->readStdErr();
  while (_pari->stdOutAvailable())
    output += _pari->readStdOut();

  ((QMainWindow*)parentWidget())->statusBar()->message("");

#ifdef _WIN32
  output.replace (QRegExp("\\r"), "");              // (only MS-Windows: remove "Carriage Return")
#endif
  output.replace (QRegExp(" *\\n* *$"), "");        //
  output.replace (QRegExp(PARI_PROMPT "\\n?"), ""); // remove the prompt
  output.replace (QRegExp("^\\n"), "");             // remove `newline` in the beginning
  output.replace (QRegExp("\\n$"), "");             // and at the end
  error.replace (QRegExp("\\r"), "");       // (only MS-Windows: remove "Carriage Return")
  error.replace (QRegExp("  \\*\\*\\*   division by zero in gdiv, gdivgs or ginv *\\n? *$"), "");
  error.replace (QRegExp(" *\\n? *$"), "");
  error.replace (QRegExp("^ *\\n* *"), "");

  if (! error.isEmpty())
    error = "<font color=" ERROR_COLOR ">" + error + "</font>";
}

void EditWindow::displayTimeAndType() {
   // show the result-type and the calculation-time in the status bar.
  _pari->writeStdIn ("##\nprint(type(%))\n1/0\n");
  while (! _pari->stdErrAvailable())
    GpController::sleep (50);
  _pari->readStdErr();
  QString info = _pari->readStdOut();
  _matrix = info.find ("t_MAT") >= 0;
  info.replace (QRegExp("^.*computed in"), tr("computed in:"));
  info.replace (QRegExp(PARI_PROMPT), "");
  info.replace (QRegExp("ms\\..*t_"), "ms, Typ: ");
  ((QMainWindow*)parentWidget())->statusBar()->message(info);
}


void EditWindow::copyInput (int p) {

  if (p >= 0) {
    int p0 = _record.find (">", p) + 1;
    if (p0 >= 0) {
      int p1 = _record.find ("</font>", p0+1);
      if (p1 >= 0) {
        QString s = _record.mid(p0, p1-p0);
        s.replace (QRegExp("&nbsp;"), " ");
        s.replace (QRegExp("<br/>"), "\n");
        s.replace (QRegExp("&lt;"), "<");
        _inputWnd->setText (s);
        _insertionPos = p;
      }
    }
  }
}

void EditWindow::up() {
  if (_insertionPos > 0)
    copyInput (_record.findRev ("<font color=" INPUT_ECHO_COLOR ">", _insertionPos-1));
}

void EditWindow::down() {
  if (_insertionPos < (int)_record.length())
    copyInput (_record.find ("<font color=" INPUT_ECHO_COLOR ">", _insertionPos+1));
}

void EditWindow::calc() {
  _inputWnd->selectAll();
  _matrix = false;
  QString pariInput = _inputWnd->text();    // --> stdin from PARI

#ifdef _WIN32
  pariInput.replace (QRegExp("\\r"), "");   // "Carriage Return" (only MS-Windows)
#endif

  QString htmlInput = pariInput;            // --> output in the protocol-window
  // newline and space for the HTML-output
  htmlInput.replace (QRegExp("<"), "&lt;"); // HTML...

  while (htmlInput.right(1) == "\n")

  htmlInput.remove (htmlInput.length()-1,1);
  htmlInput.replace (QRegExp("\\n"), "<br/>");
  htmlInput.replace (QRegExp(" "), "&nbsp;");

  QString append = _firstInput ? "" : "<hr/>";
  _firstInput = false;
  append += "<font color=" INPUT_ECHO_COLOR ">" + htmlInput + "</font><br/>";
 _record += append;

  QString output, error;
  dispatch (pariInput, output, error);

  if (! error.isEmpty())
    output += error;
  else if (output.find(QRegExp("^%\\d+")) >= 0)
    displayTimeAndType();
  if (_matrix)
    formatMatrix (output);
  output.replace (QRegExp("  "), " &nbsp;"); // space-char for HTML
  output.replace (QRegExp("\\n"), "<br/>");  // newline for HTML
  appendRecord (output);

  _insertionPos = _record.findRev ("<font color=" INPUT_ECHO_COLOR ">");
}

#if (!defined KEIN_PATCH_WEGEN_QT_FEHLER) //no patch, because of an Qt-error
void EditWindow::scrollDown() {
  _recordWnd->setContentsPos (0, 0x10000);
  _inputWnd->setFocus();
}
#endif


void EditWindow::open() {
  QString fileName = QFileDialog::getOpenFileName (QString::null, "*.gp", this, "", tr("Open Pari-GP file"));
  if (! fileName.isEmpty()) {
    QString tempFile = _pari->programPath();

#ifdef _WIN32
    tempFile += "~temp.gp";
    GpController::copyFile (fileName, tempFile);
    _pari->writeStdIn ("read(\"~temp.gp\")\n1/0\n");
#endif

#ifdef _LINUX
    tempFile = "/tmp/pariguide_tmp.gp";
    GpController::copyFile (fileName, tempFile);
//    read((open("/tmp/pariguide_tmp.gp", O_RDONLY)), b, MAX_BUFF);
//    _pari->writeStdIn (b + "\n1/0\n");
#endif
   
    while (! _pari->stdErrAvailable())
      GpController::sleep (50);
    while (_pari->stdOutAvailable())
      _pari->readStdOut();
    while (_pari->stdErrAvailable())
      _pari->readStdErr();

    GpController::deleteFile (tempFile);
  }
}

void EditWindow::save() {
  QString fileName = QFileDialog::getSaveFileName (QString::null, "*.html", this, "", tr("Save protocol as HTML file"));
  if (! fileName.isEmpty()) {
    if (fileName.right(5) != ".html")
      fileName += ".html";
    QFile f(fileName);
    f.open (IO_WriteOnly);
    char head[] = "<html><head><title>pariGUIde protocol</title></head><body>\n"
                  "<!--$|" INPUT_ECHO_COLOR "|$-->\n";
    char tail[] = "</body></html>";
    f.writeBlock (head, strlen(head));
    int length = _record.length();
    const char* body = (const char*)_record;
    int p0 = 0, p1;
    bool cont = true;
    while (cont) {
      p1 = _record.find (QRegExp("<[bh]r/>"), p0);
      if (p1 < 0) {
        cont = false;
        p1 = length;
      }
      else p1 += 5;
      //f.writeBlock ((const char*) _record, _record.length());
      f.writeBlock (body+p0, p1-p0);
      f.writeBlock ("\n", 1);
      p0 = p1;
    }
    f.writeBlock (tail, strlen(tail));
    f.close();
  }
}

bool EditWindow::openDemo() {
  QString fileName = QFileDialog::getOpenFileName (QString::null, "*.html", this, "", tr("Open protocol as presentation"));
  if (fileName.isEmpty())
    return false;
  else {
    _presentation = "";
    QFile f(fileName);
    f.open (IO_ReadOnly);
    QString s;
    int nRead = 1;
    while (nRead > 0) {
      nRead = f.readLine (s, 4096);
      _presentation += s;
    }
    f.close();

    QString inputColor = INPUT_ECHO_COLOR;
    // <!--$|color="red"|$-->
    int p0 = _presentation.find("<!--$|");
    if (p0 >= 0) {
     // if the color change in the future
      p0 += 6;
      int p1 = _presentation.find("|$-->", p0);
      inputColor = _presentation.mid(p0, p1-p0);
    }
    _presentation.replace (QRegExp("<font color=" + inputColor + ">"), "");
    _presentation.replace (QRegExp("<font color=" INPUT_ECHO_COLOR ">"), "");
    _presentation.replace (QRegExp("</font>"), "");
    _presentation.replace (QRegExp("^[^]*"), "");
    _presentation.replace (QRegExp("[^]*"), "");
    _presentation.replace (QRegExp("[^]*$"), "");
    _presentation.replace (QRegExp("[]"), "");
    _presentation.replace (QRegExp("&nbsp;"), " ");
    _presentation.replace (QRegExp("&lt;"), "<");
    _presentation.replace (QRegExp("[\\r\\n]"), "");
    _presentation.replace (QRegExp("<br/>"), "\n");
    _presentation.replace (QRegExp("+"), "");
    _presentation.replace (QRegExp("^"), "");
    demoStep();
    return true;
  }
}

bool EditWindow::demoStep() {
  bool more = true;
  if (_presentation.isEmpty())
    return false;
  int pos = _presentation.find('');
  QString s;
  if (pos >= 0) {
    s = _presentation.left(pos);
    _presentation.remove (0, pos+1);
  }
  else {
    s = _presentation;
    _presentation = "";
    more = false;
  }
  _inputWnd->setText (s);
  return more;
}

void EditWindow::stop() { // not complete, TODO
  GpController::setBreakHotkeyPressed(true);
}

void EditWindow::selectFont() {
  bool ok;
  QFont f = QFontDialog::getFont (&ok, _font, this);
  if (ok) {
    _font = f;
    _inputWnd->setFont (_font);
    _recordWnd->setFont (_font);
    _recordWnd->setText (_record); // refresh the font
    scrollToEndOfRecord();
  }
}

void EditWindow::setPrecision() {
  bool ok = false;

  //get old precision
  _pari->writeStdIn ("\\p\n1/0\n");
  while (! _pari->stdErrAvailable())
    GpController::sleep (50);
  _pari->readStdErr();
  QString old = _pari->readStdOut();
  old.replace (QRegExp(PARI_PROMPT), "");
  old.replace (QRegExp("[a-zA-Z]"), "");
  old.replace (QRegExp("\\s"), "");
  old.replace (QRegExp("="), "");
  int oldPrecision = old.toInt ( &ok, 10);

  int n = QInputDialog::getInteger (tr("Computational accuracy"),
                                    tr("significante digits by real numbers:"),
                                    _realPrecision, 2, 1000, 1, &ok, this);

  if (ok && (oldPrecision != n)) { //prevent deadlock under linux
    _realPrecision = n;
     char s[80];

    sprintf (s, "\\p%d\n1/0\n", n);
    _pari->writeStdIn (s);
    while (! _pari->stdErrAvailable())
      GpController::sleep (50);
    _pari->readStdErr();
    QString info = _pari->readStdOut(); // e.g. "realprecision = 57 significant digits (55 digits displayed)"

    info.replace (QRegExp(PARI_PROMPT), "");
    info.replace (QRegExp("^ *r"), "r");
    ((QMainWindow*)parentWidget())->statusBar()->message(info);
  }
}

void EditWindow::setPrecision (int n) {
  _realPrecision = n;
  char s[80];
  sprintf (s, "\\p%d\n1/0\n", n);
  _pari->writeStdIn (s);
  while (! _pari->stdErrAvailable())
    GpController::sleep (50);
  _pari->readStdErr();
  _pari->readStdOut();
}

void EditWindow::insertSum() {
  SumDialog dlg (this, "", true);
  if (dlg.exec()) {
    QString s = "sum (";
    s += dlg.editVar->text() + "=";
    s += dlg.editFrom->text() + ",";
    s += dlg.editTo->text() + ", ";
    s += dlg.editTerm->text() + ")";
    _inputWnd->insert (s);
  }
}

void EditWindow::insertProd() {
  ProdDialog dlg (this, "", true);
  if (dlg.exec()) {

    QString s = "prod (";
    s += dlg.editVar->text() + "=";
    s += dlg.editFrom->text() + ",";
    s += dlg.editTo->text() + ", ";
    s += dlg.editTerm->text() + ")";
    _inputWnd->insert (s);
  }
}

void EditWindow::insertIf() {
  _inputWnd->insert (""); // (gets the cursorposition after deletion of the selection)
  int line, col;
  _inputWnd->getCursorPosition (&line, &col);
  QString s = "if (#,\n  #\n,\\\\else\n  #\n)";
  char* t = new char[col+1];
  memset (t, ' ', col); t[col] = '\0';
  s.replace (QRegExp(""), t);
  delete[] t;
  _inputWnd->insert (s);
  _inputWnd->setCursorPosition (line, col+4);
}

void EditWindow::insertMatrix() {
  MatrixDialog dlg (this, "", true);
  if (dlg.exec()) {
    QString s = "matrix (";
    s += dlg.editN->text() + ",";
    s += dlg.editM->text() + ", ";
    s += dlg.editI->text() + ",";
    s += dlg.editK->text() + ", ";
    s += dlg.editAik->text() + ")";
    _inputWnd->insert (s);
  }
}

void EditWindow::insertTable() {

  static QString sX0   = "1";
  static QString sN    = "10";
  static QString sStep = "1";
  static QString sFn1 = "x^2";
  static QString sFn2 = "2^x";
  static QString sFn3 = "x^x";
  static bool s_vert  = true;

  TableDlg dlg (this, "", true);
  dlg.editX0->setText (sX0);
  dlg.editN->setText (sN);
  dlg.editStep->setText (sStep);
  dlg.editFn1->setText (sFn1);
  dlg.editFn2->setText (sFn2);
  dlg.editFn3->setText (sFn3);
  dlg.checkVertical->setChecked (s_vert);

  if (dlg.exec()) {
    sX0  = dlg.editX0->text();
    sN   = dlg.editN->text();
    sStep= dlg.editStep->text();
    sFn1 = dlg.editFn1->text();
    sFn2 = dlg.editFn2->text();

    sFn3 = dlg.editFn3->text();

    QString term = "j-2";
    if (sStep != "1")
      term = "(" + term + ")*" + sStep;
    if (sX0 != "0")
      term = sX0 + "+" + term;

    QString fn1 = sFn1;
    QString fn2 = sFn2;
    QString fn3 = sFn3;
    int nFunctions = 1;
    if (! fn2.isEmpty()) {
      nFunctions = 2;
      if (! fn3.isEmpty()) nFunctions = 3;
    }
    QString s;
    switch (nFunctions) {
    case 1: s = "print(matrix(2,N,2,I,"
                "if(i==1,"   "if(j==1,\"x\",A),"
                             "if(j==1,T1,F1))))";   break;

    case 2: s = "print(matrix(3,N,3,I,"
                "if(i==1,"   "if(j==1,\"x\",A),"
                "if(i==2,"   "if(j==1,T1,F1),"
                             "if(j==1,T2,F2)))))";  break;

    case 3: s = "print(matrix(4,N,4,I,"
                "if(i==1,"   "if(j==1,\"x\",A),"
                "if(i==2,"   "if(j==1,T1,F1),"
                "if(i==3,"   "if(j==1,T2,F2),"
                             "if(j==1,T3,F3))))))"; break;
    }
    if (dlg.checkVertical->isChecked()) {
      s.replace (QRegExp("..N"), "N");
      s.replace (QRegExp(",I,"), ",j,i,");
    }
    else {
      s.replace (QRegExp("N.."), "N");
      s.replace (QRegExp(",I,"), ",i,j,");
    }
    s.replace (QRegExp("N"), sN+"+1");
    s.replace (QRegExp("A"), term);
    s.replace (QRegExp("T1"), "\"" + fn1 + "\"");
    s.replace (QRegExp("T2"), "\"" + fn2 + "\"");
    s.replace (QRegExp("T3"), "\"" + fn3 + "\"");
    term = "(" + term + ")";
    fn1.replace (QRegExp("x"), term);  s.replace (QRegExp("F1"), fn1);
    fn2.replace (QRegExp("x"), term);  s.replace (QRegExp("F2"), fn2);
    fn3.replace (QRegExp("x"), term);  s.replace (QRegExp("F3"), fn3);

    _record += "<hr/>";

    QString output, error;
    dispatch (s, output, error);

    if (! error.isEmpty())
      output += error;
    else if (output.find(QRegExp("^%\\d+")) >= 0)
      displayTimeAndType();

    output.replace (QRegExp("\""), "<b>");
    output.replace (QRegExp("\""), "</b>");

    formatMatrix (output, true);
    output.replace (QRegExp("  "), " &nbsp;"); // space-char for HTML
    output.replace (QRegExp("\\n"), "<br/>");  // newline for HTML
    appendRecord (output);
  }
}

void EditWindow::insertPlot() {
  _recordWnd->setText (_record);

  static QString sX0 = "0";
  static QString sX1 = "2*Pi";
  static QString sFn = "sin (x)";

  PlotDialog dlg (this, "", true);
  dlg.editX0->setText (sX0);
  dlg.editX1->setText (sX1);
  dlg.editFn->setText (sFn);
  if (! dlg.exec())
    return;

  sX0 = dlg.editX0->text();
  sX1 = dlg.editX1->text();
  sFn = dlg.editFn->text();

  int precision = _realPrecision;
  setPrecision (5); // minimize the data transfer

  QString output, error;

  QString sValX0, sValX1;
  double valX0, valX1;

  // analyse X0, X1 (formula --> literalconstant)
  for (int j=0; j<2; j++) {
    QString s = "print(" + ((j==0) ? sX0 : sX1) + ")";
    output = "";
    error = "";
    dispatch (s, output, error);

    if (! error.isEmpty()) {
      appendRecord (error);

      setPrecision (precision);
      return;
    }
    if (j==0) sValX0 = output;
    else      sValX1 = output;

  }
  valX0 = sValX0.toDouble();
  valX1 = sValX1.toDouble();

  int nStrokes = 200;
  QString sStrokes; sStrokes.setNum (nStrokes);

  sStrokes += ".0";
  QString sStrokes1; sStrokes1.setNum (nStrokes+1);
  QString sFnReplace = sFn;

  QString sArg = "(" + sValX0 + "+(i-1)*((" + sValX1 + ")-(" + sValX0 + "))/" + sStrokes + ")";
  sFnReplace.replace (QRegExp("x"), "()");
  QString s = "print(vector("+ sStrokes1 + ",i," + sFnReplace + "))";
  s.replace (QRegExp(""), sArg);

  output = "";
  dispatch (s, output, error);

  if (! error.isEmpty()) {
    appendRecord (error);
    setPrecision (precision);
    return;
  }

  _plotWnd->beginPlot (valX0, valX1, nStrokes+1);

  int pos0 = 0, pos1 = 0;
  for (int i=0; i<=nStrokes; i++) {
    // localise the next value of the function in the vector
    pos0 = output.find (QRegExp("[-.\\dE]"), pos1);
    pos1 = output.find (QRegExp("[^-.\\dE]"), pos0+1);
    double d = output.mid(pos0,pos1-pos0).toFloat();
    _plotWnd->addPlotValue (d);
  }
  _plotWnd->endPlot();
  _tabView->showPage(_plotWnd);

  setPrecision (precision);
}


void EditWindow::insertInt() {
  IntDialog dlg(this, "", true);
  if (dlg.exec()) {
      QString tmp;
      int flag = dlg.flagType->currentItem();

      QString s = "intnum(";  // e. g. QString b = "intnum (x = 10, 20, x^2,{flag = 0})";
      s += dlg.intVar->text() + "=";
      s += dlg.fromA->text() + ",";
      s += dlg.toB->text() + ",";
      s += dlg.fktnExpr->text() + ",";
      tmp.sprintf("{flag=%d})", flag);
      s += tmp;
      _inputWnd->insert(s);
  }
}
