/**************************************************************************
 *  Copyright (c) 2002  Hartmut Ring, Universitaet Siegen                 *
 *  http://www.math.uni-siegen.de/ring                                    *
 *                                                                        *
 *  This file is part of the pariGUIde package.                           *
 *                                                                        *
 *  pariGUIde is free software; you can redistribute it and/or modify it  *
 *  under the terms of the GNU General Public License (GPL)               *
 *  as published by the Free Software Foundation:                         *
 *  http://www.opensource.org/licenses/gpl-license.html                   *
 *************************************************************************/
// plotview.cpp

#include <math.h>
#include <float.h>

#include "plotview.h"
#include <qpainter.h>
#include <qfontmetrics.h>


PlotView::PlotView (QWidget* parent/*=0*/)
: QWidget (parent),
  _nValues (0),
  _state (NONE),
  _y (NULL)
{
  setBackgroundColor (white);
  //setBackgroundColor (QColor(240,240,240));
  //setMouseTracking(true); // sonst nur bei gedrueckter Maus
}

PlotView::~PlotView () {
  if (_y) delete[] _y;
}


void PlotView::beginPlot (double xMin, double xMax, int nValues) {
  if (nValues != _nValues) {
    if (_y) delete[] _y;
    _y = new double[nValues];
    _nValues = nValues;
  }
  _iValue = 0;
  _xMin = xMin;
  _xMax = xMax;
  _yMin = DBL_MAX;
  _yMax = DBL_MIN;
}

bool PlotView::addPlotValue (double y) {
  if (_iValue >= _nValues)
    return false;
  _y[_iValue++] = y;
  if (y < _yMin) _yMin = y;
  if (y > _yMax) _yMax = y;
  return true;
}

void PlotView::endPlot () {
  _state = PLOT;
  // y-Intervall nach auen auf 2 signifikante Stellen runden
  round (_yMin, false, 2);
  round (_yMax, true,  2);
  repaint();
}

void PlotView::round (double& y, bool up, int digits/*=2*/) {
  if (y != 0) {
    bool neg = y < 0;
    if (neg) y = -y;
    double factor = pow (10.0, floor(log10(y))-digits+1);
    y /= factor;
    if (neg) y = -y;
    y = (up ? ceil (y) : floor (y)) * factor;
  }
}

static double gridBase (int d, double scale) {
  // kleinste Zehnerpotenz x, so dass scale * x >= d
  return  pow (10, ceil(log10(d/scale)));
}

void PlotView::drawGrid (QPainter& p, int d, int e) {
  // senkrechte Gitterlinien
  double dxGrid = e * gridBase (d, _scaleX);
  double xGrid = dxGrid * ceil(_xMin/dxGrid);
  while (xGrid <= _xMax) {
    int x0 = _left + (xGrid-_xMin) * _scaleX;
    p.drawLine (x0, _top, x0, _bottom);
    xGrid += dxGrid;
  }
  // waagerechte Gitterlinien
  double dyGrid = e * gridBase (d, _scaleY);
  double yGrid = dyGrid * floor(_yMin/dyGrid);
  while (yGrid <= _yMax) {
    int y0 = _bottom - (yGrid-_yMin) * _scaleY;
    p.drawLine (_left, y0, _right, y0);
    yGrid += dyGrid;
  }
}

void PlotView::paintEvent (QPaintEvent* ev) {
  QPainter p(this);
  ASSERT (p.rasterOp() == Qt::CopyROP);

  QString sxMin; sxMin.setNum (_xMin);
  QString sxMax; sxMax.setNum (_xMax);
  QString syMin; syMin.setNum (_yMin);
  QString syMax; syMax.setNum (_yMax);

  QFontMetrics fm = p.fontMetrics();
  _fontH = fm.height();
  _left = fm.width (syMin);
  if (fm.width (syMax) > _left)
    _left = fm.width (syMax);
  _left += 4;

  _top  =  8;
  _right =  width() - 8;
  _bottom = height() - _fontH;

  _scaleX = (_right - _left) / (_xMax - _xMin);
  _scaleY = (_bottom - _top) / (_yMax - _yMin);

  int w = (_right-_left)/2;

  switch (_state) {
  case PLOT: {
    // Gitterlinien
    p.setPen (QColor(240,240,240));  drawGrid (p, 4, 1);
    p.setPen (QColor(224,224,224));  drawGrid (p, 4, 5);
    p.setPen (QColor(192,192,192));  drawGrid (p, 4, 10);

    // Koordinatenachsen
    p.setPen (Qt::blue);
    if (_yMin <= 0.0 && _yMax >= 0.0) {
      int y0 = _bottom + _yMin * _scaleY;
      p.drawLine (0, y0, width(), y0);
    }
    if (_xMin <= 0.0 && _xMax >= 0.0) {
      int x0 = _left  -_xMin * _scaleX;
      p.drawLine (x0, 0, x0, height());
    }
#ifdef RAHMEN_ZEICHNEN
    p.setPen (Qt::green);
    p.drawRect (_left, _top, _right-_left, _bottom-_top);
#endif

    // Beschriftung
    p.setPen (Qt::black);

    p.drawText (0, _top,           _left-4, _fontH,  Qt::AlignRight, syMax);
    p.drawText (0, _bottom-_fontH, _left-4, _fontH,  Qt::AlignRight, syMin);

    p.drawText (_left+2,    _bottom, w-2,   _fontH,  Qt::AlignLeft,  sxMin);
    p.drawText (_right-w-2, _bottom, w-2,   _fontH,  Qt::AlignRight, sxMax);

    // Plot
    p.moveTo (_left, _bottom - (_y[0]-_yMin) *_scaleY);
    for (int i=1; i<_nValues; i++)
      p.lineTo (_left+i*(double)(_right-_left)/(_nValues-1), _bottom - (_y[i]-_yMin)*_scaleY);
    break;
  }
  case BAR:
    break;
  default:
    p.setPen (Qt::black);
    p.drawText (0, (height()-_fontH)/2, width(), _fontH,
                Qt::AlignCenter,
                tr("Please define a plot (Menu: Paste --> Graph)"));
  }


}
